import logging
import pathlib
import shutil
from collections import defaultdict
from typing import Callable

from jinja2 import Environment, FileSystemLoader

from .collection import Collection
from .page import Page


class Site:
    """
    The site stores your pages and collections to be rendered.

    Attributes:
        output_path: str to write rendered content. **Default**: `output`
        static: Output str Path for the static folder. This will get copied to the output folder. **Default**: `static`
        site_vars: Vars that will be passed into the render functions

            Default `site_vars`:

            - SITE_TITLE: "Untitled Site"
            - SITE_URL: "http://example.com"
    """

    output_path: str = "output"
    static_path: str = "static"
    # TODO: #74 Should this be called from a config file for easier testing?
    site_vars: dict = {
        "SITE_TITLE": "Untitled Site",
        "SITE_URL": "https://example.com",  # TODO: #73 Make this http://localhost:8000
    }
    plugins: dict[str, Page] | None = None

    def __init__(
        self,
    ) -> None:
        self.route_list: defaultdict = defaultdict(list)
        self.subcollections: defaultdict = defaultdict(lambda: {"pages": []})

    @property
    def engine(self) -> Environment:
        env = Environment(loader=FileSystemLoader("templates"))
        env.globals = self.site_vars
        return env

    def add_to_route_list(self, page: Page) -> None:
        """Add a page to the route list"""
        self.route_list[page.url_for] = page

    def collection(self, collection: Collection):
        """Create the pages in the collection including the archive"""
        _collection = collection(engine=self.engine, **self.site_vars)
        logging.info("Adding Collection: %s", _collection.__class__.__name__)

        for page in _collection.pages:
            logging.debug("Adding Page: %s", page.__class__.__name__)
            self.add_to_route_list(page)

        for archive in _collection.archives:
            logging.debug("Adding Archive: %s", archive.__class__.__name__)
            self.add_to_route_list(archive)

        if feed := (getattr(_collection, "_feed", None)):
            self.add_to_route_list(feed)

    def page(self, page: Page) -> None:
        """Create a Page object and add it to self.routes"""
        logging.info("Adding Page: %s", page.__class__.__name__)
        _page = page(self.engine, **self.site_vars)
        self.add_to_route_list(_page)

    def render_static(self, directory) -> None:
        """Copies a Static Directory to the output folder"""

    def render_output(self, route, page):
        """writes the page object to disk"""
        if page._extension == ".xml":
            logging.debug("%s, %s", page.content, page.pages)
        path = (
            pathlib.Path(self.output_path)
            / pathlib.Path(route)
            / pathlib.Path(page.url)
        )
        path.parent.mkdir(parents=True, exist_ok=True)
        return path.write_text(page._render_content())

    def build_subcollections(self, page) -> None:
        if subcollections := getattr(page, "subcollections", []):
            logging.debug("Adding subcollections: %s", subcollections)

            for attr in subcollections:
                logging.debug("Adding attr: %s", attr)

                for page_attr in getattr(page, attr, []):
                    logging.debug("Adding page_attr: %s", page_attr)
                    self.subcollections[page_attr]
                    self.subcollections[page_attr]["pages"].append(page)
                    self.subcollections[page_attr]["route"] = attr

                    if "template" not in self.subcollections[page_attr]:
                        self.subcollections[page_attr][
                            "template"
                        ] = page.subcollection_template

    def render(self, clean=False) -> None:
        """Render all pages and collections"""

        if clean:
            shutil.rmtree(self.path, ignore_errors=True)

        # Parse Route List
        for page in self.route_list.values():
            self.build_subcollections(page)

            for route in page.routes:
                self.render_output(route, page)

        # Parse SubCollection
        for tag, subcollection in self.subcollections.items():
            page = Page(
                self.engine,
                title=tag,
                template=subcollection["template"],
                pages=subcollection["pages"],
            )
            self.render_output(
                Path(page.routes[0]).joinpath(subcollection["route"]), page
            )

        if path := (pathlib.Path(self.static_path)).exists():
            self.render_static(path)



import itertools
import typing
from collections import defaultdict
from pathlib import Path

import jinja2
from more_itertools import chunked

from .page import Page


class Archive(Page):
    """Custom Page object used to make archive pages"""

    def __init__(self, /, pages: list, template: str, **kwargs) -> None:
        """Create a `Page` object for the pages in the collection"""
        super().__init__(**kwargs)
        self.pages = pages
        self.template = template


def gen_collection(
    pages: typing.Iterable[Page],
    template: jinja2.Template,
    title: str,
    items_per_page: typing.Optional[int] = None,
    routes: list[Path | str] = [],
    collection_vars: dict = {},
) -> list[Archive]:
    """Returns a list of Archive pages containing the pages of data for each archive."""

    if not items_per_page:
        return [
            Archive(
                engine=template.environment,
                pages=pages,
                template=template,
                title=title,
                routes=routes,
                **collection_vars,
            )
        ]

    page_chunks = chunked(pages, items_per_page)

    pages = [
        Archive(
            engine=template.environment,
            pages=pages,
            template=template,
            slug=f"{title}_{i}",
            title=title,
            routes=routes,
            **collection_vars,
        )
        for i, pages in enumerate(page_chunks)
    ]
    return pages


SubCollection = {str, list[Page]}


class Collection:
    """Collection objects serve as a way to quickly process pages that have a
    LARGE portion of content that is similar or file driven.

    The most common form of collection would be a Blog, but can also be
    static pages that have their content stored in a dedicated file.

    Currently, collections must come from a content_path and all be the same
    content type.

    Example::

        from render_engine import Collection

        @site.register_collection
        class BasicCollection(Collection):
            pass
    """

    content_path: Path
    content_type: Page = Page
    template: typing.Optional[str] = None
    includes: list[str] = ["*.md", "*.html"]
    markdown_extras = ["fenced-code-blocks", "footnotes"]
    items_per_page: typing.Optional[int] = None
    sort_by: str = "title"
    sort_reverse: bool = False
    has_archive: bool = False
    archive_template: typing.Optional[str] = None
    subcollections: list[str | list[str]] = None
    routes = ["./"]

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)

        if not hasattr(self, "title"):
            self.title = self.__class__.__name__

        if any([self.items_per_page, self.archive_template]):
            self.has_archive is True
            self.archive_template = self.engine.get_template(self.archive_template)

            if hasattr(self, "subcollections") and not getattr(
                self, "subcollections_template", None
            ):
                self.subcollection_template = self.archive_template

        self.routes = [Path(route) for route in self.routes]

    @property
    def collection_vars(self):
        """
        Creates Collection Vars to Pass into template.
        """
        return {key.upper(): val for key, val in vars(self).items()}

    @property
    def pages(self):
        if Path(self.content_path).is_dir():
            pages = self._pages(self.content_type, routes=self.routes)

            return pages
        else:
            raise ValueError("invalid Path: %s}" % Path)

    def _pages(self, content_type: Page, **kwargs) -> list[Page]:
        page_groups = map(
            lambda pattern: Path(self.content_path).glob(pattern), self.includes
        )
        return [
            content_type(
                engine=self.engine,
                content_path=page_path,
                template=self.template,
                subcollections=getattr(self, "subcollections", []),
                subcollection_template=getattr(self, "subcollection_template", []),
                **self.collection_vars,
                **kwargs,
            )
            for page_path in itertools.chain.from_iterable(page_groups)
        ]

    @property
    def sorted_pages(self):
        return sorted(
            self.pages,
            key=lambda page: getattr(page, self.sort_by),
            reverse=self.sort_reverse,
        )

    @property
    def archives(self) -> list[Archive]:
        """Returns a list of Archive pages containing the pages of data for each archive."""
        if self.has_archive:
            return gen_collection(
                pages=self.sorted_pages,
                template=self.archive_template,
                title=self.title,
                items_per_page=self.items_per_page,
                routes=self.routes,
                collection_vars=self.collection_vars,
            )
        return ()

    # def render_feed(self, feed_type: RSSFeed, **kwargs) -> RSSFeed:
    #     return feed_type(pages=self.pages, **kwargs)

    def __repr__(self):
        return f"{self}: {__class__.__name__}"

    def __str__(self):
        return f"{self}: {__class__.__name__}"

    def __iter__(self):
        return iter(self.pages)


def render_archives(archive, **kwargs) -> list[Archive]:
    return [archive.render(pages=archive.pages, **kwargs) for archive in archive]



"""
The Feeds Logic That Makes Up RSS and ATOM FeedTypes.
This is the base files and should only contain the params identified by the
standards defined.
RSS: http://www.rssboard.org/rss-specification
JSON: https://jsonfeed.org/version/1
"""

from collections import namedtuple
from datetime import datetime

import jinja2
from jinja2 import Template, select_autoescape

from .loaders import render_engine_templates_loader
from .page import Page


def to_pub_date(value: datetime):
    """
    Parse information from the given class object.
    """
    return value.to_rfc2822_string()


rss_feed_engine = jinja2.Environment(
    loader=render_engine_templates_loader,
    autoescape=select_autoescape(enabled_extensions=("rss")),
    trim_blocks=True,
)
rss_feed_engine.filters["to_pub_date"] = to_pub_date


class RSSFeed(Page):
    """The RSS Feed Component of an Archive Object"""

    engine = rss_feed_engine
    template = "rss2.0.xml"
    extension: str = "rss"

    def __init__(self, title, pages, **kwargs):
        self.title = getattr(self, "title", title)
        self.pages = pages
        super().__init__(**kwargs)



from dataclasses import dataclass, field


@dataclass
class Link:
    """
    An opinionated format to reference links. Great for creating links that you will pass into multiple page objects.

    Jinja will pass the string representation of the link to the template.

    .. code-block:: python

        link = Link(
                name="Render Engine",
                url="https://render-engine.site",
                meta={"class"="link-class", "id"="custom-link-id"}
                )

        # When rendered in a template, this will be:

        <a href="https://render-engine.site" class="link-class" id="custom-link-id">Render Engine</a>
    """

    text: str = field(default_factory=str)
    url: str = field(default="#")
    meta: dict[str, str] = field(default_factory=dict)
    """Meta variables translate to attributes on the html element"""

    def __str__(self):
        if self.meta:
            attrs = " ".join([f'{k}="{v}"' for k, v in self.meta.items()])
            starting_path = f'<a href="{self.url}" {attrs}>'

        else:
            starting_path = f'<a href="{self.url}">'

        return f"{starting_path}{self.text}</a>"


@dataclass
class Image(Link):
    """
    Link object formatted as an image

    .. code-block:: python

        myImage = Link(
            text="Render Engine",
            url="https://render-engine.site",
            meta={"class"="link-class", "id"="custom-link-id"}
            )

        # When rendered in a template

        <img src="https://render-engine.site" alt="example image" id="custom-link-id" class="link-class" />
    """

    def __str__(self):
        """prints the image point"""

        if self.meta:
            attrs = " ".join([f'{k}="{v}"' for k, v in self.meta.items()])
            return f'<img src="{self.url}" alt="{self.text}" {attrs} />'

        else:
            return f'<img src="{self.url}" alt="{self.text}" />'



import logging
from pathlib import Path
from typing import Generator, Optional

import frontmatter
import jinja2
from markdown2 import markdown
from slugify import slugify

_route = Path | str


class Page:
    """The base object used to make web pages.
    Pages can be rendered directly from a template or generated from a file.

    !!! note

        Not all attributes are defined by default (those that are marked *optional*) but
        will be checked for in other areas of the code.

    When you create a page, you can specify variables that will be passed into rendering template.

    Attributes:
        
    """

    markdown_extras: list[str] = ["fenced-code-blocks", "footnotes"]
    """Plugins that will be used with the markdown parser (default parser is [Markdown2](https://github.com/trentm/python-markdown2)).
     You can see a list of all the plugins [here](https://github.com/trentm/python-markdown2/wiki/Extras).

     The default plugins fenced-code-blocks and footnotes provide a way to add code blocks and footnotes to your markdown.
     """

    markdown: str | None = None
    """This is base markdown that will be used to render the page.

    !!! warning

        This will be overwritten if a `content_path` is provided.
    """

    content_path: _route | None = None
    """
    The path to the file that will be used to generate the page.

    !!! note

        This overrides the `markdown` attribute.
    """

    extension: str = ".html"
    """Extension to use for the rendered page output."""

    reference: str = "slug"

    routes: list[_route] = ["./"]

    template: str | None = None

    def __init__(self, engine: jinja2.Environment | None = None, **kwargs) -> None:
        if not hasattr(self, "engine"):
            self.engine = engine

        for key, val in kwargs.items():
            setattr(self, key, val)

        if self.markdown and self.content_path is not None:
            logging.warning(
                "both `Page.markdown` and `content_path` selected. the content from `content_path` will be used."
            )

        if self.content_path:
            post = frontmatter.load(Path(self.content_path))
            valid_attrs, self.markdown = post.metadata, post.content
            logging.info(f"content_path found! {valid_attrs=}, {self.markdown=}")

            for name, value in valid_attrs.items():
                # comma delimit attributes.
                if name.lower() in getattr(self, "list_attrs", []):
                    value = [attrval.lower() for attrval in value.split(", ")]

                setattr(self, name.lower(), value)

        if not hasattr(self, "title"):
            self.title = self.__class__.__name__
            logging.info(f"No Title. Assigning {self.title=}")

        if not hasattr(self, "slug"):
            logging.info(f"No slug. Will slugify {self.title=}")
            self.slug = self.title  # Will Slugify in Next Step

        if not self.routes:
            self.routes = []

        self.slug = slugify(self.slug)

    @property
    def url(self) -> Path:
        """The first route and the slug of the page."""
        return f"{self.slug}{self._extension}"

    @property
    def _extension(self) -> str:
        """Ensures consistency on extension"""
        if not self.extension.startswith("."):
            return f".{self.extension}"
        else:
            return self.extension

    @property
    def content(self) -> Optional[str]:
        """html = rendered HTML (not marked up).
        Is `None` if `content == None`
        This is referred to as `content` because it is intended to be applied in the jinja template as {{content}}.
        When referring to the raw content in the Page object use `markdown`.
        """

        if not self.markdown:
            return None

        return markdown(self.markdown, extras=self.markdown_extras)

    @property
    def url_for(self):
        """Returns the url for the page"""
        return self.slug

    def __str__(self):
        return self.slug

    def __repr__(self) -> str:
        return f"<Page {self.title}>"

    def _render_content(self, **kwargs) -> str:
        template = None

        if self.template:
            logging.debug(f"Using %s for %s", self.template, self)
            template = self.engine.get_template(self.template)

        if template:
            if self.content:
                logging.debug(
                    "content found. rendering with content: %s, %s, %s",
                    self.content,
                    self.__dict__,
                    kwargs,
                )
                return template.render(
                    content=self.content, **{**self.__dict__, **kwargs}
                )

            else:
                logging.debug(
                    "No content found. rendering with content: %s, %s",
                    self.__dict__,
                    kwargs,
                )
                return template.render(**{**self.__dict__, **kwargs})

        elif self.content:
            return self.content

        else:
            raise ValueError(f"{self=} must have either content or template")

    def __iter__(self):
        """Good for getting the route objects"""
        yield from self.render()

    def render(
        self, *, engine=None, **kwargs
    ) -> Generator[dict[_route, "Page"], None, None]:
        """Build the route based on content instructions"""

        for route in self.routes:
            yield {str("Path(_route) / self.url"): self}



from .blog import Blog
from .collection import Collection
from .page import Page

# TODO: #70 site is a reserved keyword in python.
# It should possibly become a new module name
from .site import Site 



import datetime
import typing

import more_itertools
import pendulum

from .collection import Collection
from .feeds import RSSFeed
from .page import Page


class BlogPost(Page):
    """Page Like object with slight modifications to work with BlogPosts."""

    list_attrs = ["tags"]
    feed = RSSFeed

    def __init__(self, **kwargs):
        """
        checks published options and accepts the first that is listed

        Attributes:
            date : pendulum.datetime
                date parsed in datetime format. usesul for sorting and things
            date_published : str
                date formated for `RSSFeed`
            date_friendly : str
                an easy to read string version of the date
        """

        super().__init__(**kwargs)
        date_published = more_itertools.first_true(
            (
                getattr(self, "date_published", None),
                getattr(self, "publish_date", None),
                getattr(self, "date", None),
            )
        )

        if isinstance(date_published, datetime.datetime):
            self.date_published = pendulum.instance(date_published).set(
                tz=pendulum.local_timezone()
            )  # TODO: fixes issue with datetimes parsed by frontmatter being converted to datetimes instead of str
        else:
            self.date_published = pendulum.parse(date_published, strict=False).set(
                tz=pendulum.local_timezone()
            )

        date_modified = more_itertools.first_true(
            (
                getattr(self, "date_modified", None),
                getattr(self, "modified_date", None),
            ),
            default=None,
        )

        if not date_modified:
            self.date_modified = self.date_published

        else:
            self.date_modified = pendulum.parse(date_modified, strict=False).set(
                tz=pendulum.local_timezone()
            )

        self.date_friendly = self.date_modified.format("MMM DD, YYYY HH:mm A")


class Blog(Collection):
    """
    Custom :py:class:`collection.Collection` class with archiving enabled, sort by `date_published` by default.

    Todos:
        - Add Support for JSON Feeds
        - Rename the archive items so they are not private
    """

    content_type: typing.Type[BlogPost] = BlogPost
    sort_reverse: bool = True
    sort_by = "date_published"
    has_archive = True

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._feed = getattr(self, "feed", RSSFeed)(
            title=f"{self.SITE_TITLE} {self.title}", pages=self.pages
        )



import pathlib
import typing

import jinja2
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn

from render_engine import Collection, Page, Site
from render_engine.loaders import render_engine_templates_loader

environment = jinja2.Environment(
    loader=render_engine_templates_loader,
    trim_blocks=True,
    lstrip_blocks=True,
)


def create_folder(*, folder, overwrite: bool) -> pathlib.Path:
    """Create a folder if it doesn't exist or if overwrite is True"""
    new_folder = pathlib.Path(folder)
    new_folder.mkdir(parents=True, exist_ok=overwrite)
    return new_folder


path = str | None


def create_app_from_template(
    site_title: str,
    site_url: str,
    site_description: path,
    site_author: path,
    static_path: path,
    output_path: path,
    collection_path: path,
) -> str:
    return environment.get_template("create_app_py.txt").render(
        site_title=site_title,
        site_url=site_url,
        site_description=site_description,
        site_author=site_author,
        output_path=output_path,
        static_path=static_path,
        collection_path=collection_path,
    )


def create_templates_folder(
    *templates,
    templates_folder_name: str,
    exists_ok: bool,
    environment: jinja2.Environment,
) -> None:
    """Create a folder for templates and optionally create an index.html file"""
    path = pathlib.Path(templates_folder_name)
    path.mkdir(
        exist_ok=exists_ok,
    )

    for template in templates:
        content = path.joinpath(template).write_text(
            environment.get_template(template).render()
        )


def update_site_vars(optional_params: dict) -> dict:
    """Remove any optional params that are None"""
    return {key: value for key, value in optional_params.items() if value}


def create_site_with_vars(
    site_title: str,
    site_url: str,
    site_description: path,
    site_author: path,
    collection_path: path,
) -> Site:
    """Create a new site from a template"""
    site = Site()
    site_vars = {
        "site_title": site_title,
        "site_url": site_url,
    }
    optional_site_vars_params = {
        "site_title": site_author,
        "site_description": site_description,
        "collections_path": collection_path,
    }
    site_vars.update(update_site_vars(optional_site_vars_params))
    site.site_vars = site_vars
    return site


def typer_app(
    site_title: str = typer.Option(
        ...,
        "--title",
        "-t",
        help="title of the site",
        prompt=True,
        rich_help_panel="Required Attributes",
        show_default=False,
    ),
    site_url: str = typer.Option(
        ...,
        "--url",
        "-u",
        help="URL for the site",
        prompt=True,
        rich_help_panel="Required Attributes",
        show_default=False,
    ),
    site_description: typing.Optional[str] = typer.Option(
        None,
        help="(Optional): Site Description",
        rich_help_panel="Optional Attributes",
    ),
    site_author: typing.Optional[str] = typer.Option(
        None,
        help="(Optional): Author of the site",
        rich_help_panel="Optional Attributes",
    ),
    output_path: typing.Optional[str] = typer.Option(
        None,
        help="custom output folder location.",
        rich_help_panel="Optional Attributes",
    ),
    static_path: typing.Optional[str] = typer.Option(
        None,
        help="custom static folder",
        rich_help_panel="Optional Attributes",
    ),
    collection_path: typing.Optional[str] = typer.Option(
        None,
        help="create your content folder in a custom location",
        rich_help_panel="Optional Attributes",
    ),
    force: bool = typer.Option(
        False,
        "--force",
        "-f",
        help="Force overwrite of existing files",
        rich_help_panel="Flags",
    ),
    run_on_complete: bool = typer.Option(
        False,
        "--run-after",
        "-x",
        help="Run the site after creating the quickstart",
        rich_help_panel="Flags",
    ),
    skip_static: bool = typer.Option(
        False,
        "--skip-static",
        "-S",
        help="Skip copying static files",
        rich_help_panel="Flags",
    ),
    skip_collection: bool = typer.Option(
        False,
        "--skip-collection",
        "-C",
        help="Skip creating the content folder and a collection",
        rich_help_panel="Flags",
    ),
    templates_path: typing.Optional[str] = typer.Option(
        "templates",
        "--templates-path",
        help="custom templates folder",
    ),
):
    """CLI for creating a new site"""
    # creating the site object and site_vars

    site = create_site_with_vars(
        site_title=site_title,
        site_url=site_url,
        site_description=site_description,
        site_author=site_author,
        collection_path=collection_path,
    )

    # add output path
    if output_path:
        site.output_path = output_path

    collection = {
        "skip": skip_collection,
        "path": collection_path or "pages",
    }

    # creating folders unless skipped
    static = {
        "skip": skip_static,
        "path": static_path or "static",
    }

    if not static["skip"]:
        static_path = create_folder(folder=static["path"], overwrite=force)

        if not static["path"] == "static":
            site.static = static_path

    if not collection["skip"]:
        collection_path = create_folder(folder=collection["path"], overwrite=force)

    # creating the app.py file from the template
    pathlib.Path("app.py").write_text(
        create_app_from_template(
            site_title=site_title,
            site_url=site_url,
            site_description=site_description,
            site_author=site_author,
            output_path=output_path,
            static_path=static_path,
            collection_path=collection_path,
        )
    )

    # Create the templates folder and the index.html file
    templates = ["index.html", "base.html", "content.html"]
    create_templates_folder(
        *templates,
        templates_folder_name="templates",
        exists_ok=force,
        environment=environment,
    )

    # Create the collection
    if not skip_collection and collection_path:
        with Progress(SpinnerColumn()) as progress:
            task = progress.add_task("Creating collection", total=1)
            pathlib.Path(collection_path).joinpath("sample_pages.md").write_text(
                environment.get_template("base_collection_path.md").render()
            )

    if run_on_complete:
        typer.echo("Running the site")
        create_app()


def create_app():
    """This is the console script entry point for 'createapp'"""
    typer.run(typer_app)
